# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

"""
EncodingType

Overview
========

The mapping between Server data-types and Python or R.
"""

from azdata.cli.commands.app.constants import RUNTIMES
from enum import Enum
from collections import OrderedDict

import json


class EncodingType(Enum):
    """
    The mapping between Server data-types and Python and R.
    """

    types = dict()
    """
    The R and Python type mapping.
    """
    types[RUNTIMES[0]] = {
        'integer': {'integer': 'integer'},
        'float': {'numeric': 'numeric'},
        'boolean': {'logical': 'logical'},
        'string': {'character': 'character'},
        'array': {'vector': 'vector'},
        'matrix': {'matrix': 'matrix'},
        'dataframe': {'dataframe': 'data.frame', 'data.frame': 'data.frame'}
    }

    types[RUNTIMES[1]] = {
        'integer': {'int': 'int'},
        'float': {'float': 'float'},
        'boolean': {'bool': 'bool'},
        'string': {'str': 'str'},
        'array': {'numpy.array': 'vector', 'array': 'vector'},        
        'matrix': {'numpy.matrix': 'matrix', 'matrix': 'matrix'},
        'dataframe': {
            'dataframe': 'data.frame',
            'data.frame': 'data.frame',
            'DataFrame': 'data.frame',
            'pandas.DataFrame': 'data.frame'
        }
    }

    @classmethod
    def cast_client_to_server_type(cls, io_type, runtime):
        """
        Map client language specific type to server type.

        :param io_type: The type to map.
        :param runtime: Language Runtime.
        :return: The server equivalent type.
        """

        if runtime not in RUNTIMES:
            raise ValueError('Runtime language not supported "{}".'.
                             format(runtime))

        for k, v in cls.types.value[runtime].items():
            if v.get(io_type):
                return v.get(io_type)

        raise ValueError('Unsupported {} type "{}".'.format(runtime, io_type))

    @classmethod
    def integer(cls, runtime):
        return list(cls.types.value[runtime].get('integer').values())[0]

    @classmethod
    def string(cls, runtime):
        return list(cls.types.value[runtime].get('string').values())[0]

    @classmethod
    def float(cls, runtime):
        return list(cls.types.value[runtime].get('float').values())[0]

    @classmethod
    def boolean(cls, runtime):
        return list(cls.types.value[runtime].get('boolean').values())[0]

    @classmethod
    def array(cls, runtime):
        return list(cls.types.value[runtime].get('array').values())[0]

    @classmethod
    def matrix(cls, runtime):
        return list(cls.types.value[runtime].get('matrix').values())[0]

    @classmethod
    def dataframe(cls, runtime):
        return list(cls.types.value[runtime].get('dataframe').values())[0]

    @classmethod
    def is_dataframe(cls, io_type):
        return io_type in [
            'dataframe',
            'data.frame',
            'pandas.DataFrame',
            'DataFrame'
        ]

    @classmethod
    def encode(cls, server_type, value):
        """
        Cast from python as `str` to server request format.

        :param server_type: The server encoding type.
        :param value: The value to be cast.
        :return: The cast valued based on server encoding type.
        """

        if value is None:
            return None

        def dataframe(x):
            return OrderedDict([(k, x[k]) for k in list(x.keys())])

        return {
            # -- r --
            'integer': lambda x: int(x),
            'numeric': lambda x: float(x),
            'character': lambda x: x,
            'logical': lambda x: True if x.lower() == 'true' else False,
            'vector': lambda x: json.loads(x),
            'matrix': lambda x: json.loads(x),
            'dataframe': lambda x: dataframe(x),
            'data.frame': lambda x: dataframe(x),
            # -- python --
            'int': lambda x: int(x),
            'float': lambda x: float(x),
            'str': lambda x: x,
            'bool': lambda x: True if x.lower() == 'true' else False,
            # TODO: revisit when server|client syncs
            'array': lambda x: json.loads(x),
            'numpy.array': lambda x: json.loads(x),
            'numpy.matrix': lambda x: json.loads(x),
            'pandas.DataFrame': lambda x: dataframe(x),
            'DataFrame': lambda x: dataframe(x),
            'list': lambda x: json.loads(x),
        }[server_type](value)

    def describe(self):
        return self.name, self.value

    def __str__(self):
        return '<EncodingType> {0}'.format(self.value)
