# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from azdata.cli.commands.app.encoding_types import EncodingType
from azdata.cli.commands.app.constants import IO_DELIM
from azdata.cli.core.util import FileUtil
from azdata.cli.core.models.parameter_definition import ParameterDefinition

import re
import json
import os


def zip_files(filename, assets, blacklist=None, flat=False):
    """
    Archives and compresses file and directory assets into ZIP format.

    :param filename: The filename of the zip.
    :param assets: A list of file-paths, directories, and dict of content to be
                   zipped.
    :param blacklist: List of files to omit from the zip archive.
    :param flat: Maintain a flat directory structure otherwise as is.
    :return: The opened zip `file` in read mode or `None` if `read` was `False`.
    """
    from zipfile import ZipFile, ZIP_DEFLATED

    try:
        blacklist = blacklist if blacklist else []
        blacklist.append(filename)

        with ZipFile(filename, 'w', ZIP_DEFLATED) as zip_file:
            for asset in assets:
                if os.path.isfile(asset):
                    if asset not in blacklist:
                        zip_file.write(asset, os.path.basename(asset))
                else:  # dirs
                    if flat:
                        for root, dirs, files in os.walk(asset):
                            for file in files:
                                if file not in blacklist:
                                    zip_file.write(os.path.join(root, file),
                                                   os.path.basename(file))
                    else:
                        for root, dirs, files in os.walk(asset):
                            for file in files:
                                if file not in blacklist:
                                    src = os.path.join(root, file)
                                    dest = os.path.join(
                                        '.',
                                        os.path.relpath(src, asset)
                                    )
                                    zip_file.write(src, dest)
        return filename
    except Exception as e:
        msg = 'Error compressing assets "{0}"\n{1}'. \
            format(type(e).__name__, str(e))
        raise Exception(msg)


def cmd_io_to_param_defs(csv, runtime):
    try:
        parameter_defs = []

        for param in csv.split(','):
            nv = param.strip().split('=')
            param_name = nv[0]
            param_type = EncodingType.cast_client_to_server_type(nv[1], runtime)
            parameter_defs.append(
                ParameterDefinition(param_name, param_type)
            )

        return parameter_defs
    except Exception:
        raise ValueError('Error parsing CSV values for I/O.')


def cmd_io_to_dict(csv, runtime):
    """
    Helper to turn cmd inputs and outputs in cvs to a `dict`.
    Example:
    `input1=str,input2=bool` --> `{'input': 'str', 'input2': bool}`
    """
    try:
        parameter_defs = {}
        csv = [] if not csv else csv

        for param in csv.split(','):
            nv = param.strip().split('=')
            param_name = nv[0]
            param_type = EncodingType.cast_client_to_server_type(nv[1], runtime)
            parameter_defs[param_name] = param_type

        return parameter_defs
    except Exception:
        raise ValueError('Error parsing CSV values for I/O.')


def cmd_inputs_to_dict(cmd_inputs, input_defs):
    """
    Helper to convert and server encode command line input values to a `dict`.
    Command line inputs take the format of:
    ```sh
    app run --inputs a=5, b=Hello, c=True, d=@/path/to/dataframe.csv
    ```
    """
    inputs = {}

    # Ex: --inputs @/path/to/file.json
    cmd_inputs = re.split(IO_DELIM, cmd_inputs) if cmd_inputs else []

    for i in cmd_inputs:
        nv = i.strip().split('=')
        input_name = nv[0].strip()
        input_value = nv[1].strip()

        if not input_defs:
            content = {}
            if FileUtil.is_lnk_path(input_value):
                content = FileUtil.read_json(FileUtil.strip_path(input_value))
            return content

        input_def = [i for i in input_defs if i['name'] == input_name]
        if not input_def:
            msg = 'Received an unexpected input argument "{}". Supported ' \
                  'input schema:\n{}'.format(input_name,
                                             json.dumps(input_defs, indent=4))
            raise ValueError(msg)

        input_type = input_def[0].get('type')

        # value = @/path/to/input/file
        if FileUtil.is_lnk_path(input_value):
            # data.frames are treated as csv|tsv
            if EncodingType.is_dataframe(input_type):
                input_value = FileUtil.get_dict_from_csv(input_value)
            else:
                input_value = FileUtil.read_text_file(input_value)

        inputs[input_name] = EncodingType.encode(input_type, input_value)

    if input_defs:
        # -- assert all inputs were passed --
        for input_def in input_defs:
            input_name = input_def.get('name')
            if not [i for i in inputs if input_name in i]:
                defs = json.dumps(input_defs, indent=4)
                raise ValueError('Expected input argument "{}". Supported input'
                                 'schema:\n{}'.format(input_name, defs))

    return inputs


def silent_remove(files):
    """
    Attempt to remove a file whether it exists or not suppressing any errors.
    """
    import shutil

    for file in files:
        try:
            if os.path.isfile(file):
                os.remove(file)
            if os.path.isdir(file):
                shutil.rmtree(file)
        except (IOError, OSError) as e:
            pass
