# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from azdata.cli.core.clients.cli_client import CliClient
from azdata.cli.core.logging import get_logger

logger = get_logger(__name__)

__all__ = ['beget']


def beget(_):
    """Client factory"""
    return ExtensionClientMixin()


class ExtensionClientMixin(CliClient):
    def __init__(self, check_auth=False, check_eula=False):
        """
        :param check_auth: Set to `True` to have this command enforce
                           authentication before running.
        :param check_auth: Set to `True` to have this command enforce eula
                           acceptance before running.
        """
        super(ExtensionClientMixin, self).__init__(
            check_auth=check_auth, check_eula=check_eula)

    @staticmethod
    def install(source=None, index_url=None, pip_proxy=None,
                pip_extra_index_urls=None):

        from azdata.cli.core.util import Pip
        from azdata.cli.core.extensions import WheelExtension
        from azdata.cli.commands.extension.exceptions import (
            ExtensionInstallError,
            ExtensionExistsError
        )
        import os

        if not source.endswith('.whl'):
            raise ValueError(
                'Unknown extension type. Only Python wheels are supported.')

        if not os.path.exists(source):
            raise ValueError(
                'Error installing source Python wheel, file does not exist:\n'
                '"{}"'.format(source))

        whl_filename = os.path.basename(source)
        ext_name = WheelExtension.get_extension_name_from_wheel(whl_filename)
        ext_path = os.path.realpath(os.path.expanduser(source))
        target = WheelExtension.get_extension_path(ext_name)

        logger.debug('Extension wheel %s', whl_filename)
        logger.debug('Extension name %s', ext_name)
        logger.debug('Extension file path %s', ext_path)

        if WheelExtension.is_installed(ext_name):
            raise ExtensionExistsError('Extension already exists "{}"'.
                                       format(ext_name))

        out = Pip().\
            target(target).\
            index_url(index_url).\
            proxy(pip_proxy).\
            extra_index_url(pip_extra_index_urls).\
            source(source).\
            install()

        if out.return_code != 0:
            raise ExtensionInstallError(out, status=out.return_code)

        return ext_name, ext_path, target

    @staticmethod
    def remove(name):
        from azdata.cli.commands.extension.exceptions import \
            ExtensionRemoveError
        from azdata.cli.core.extensions import WheelExtension, PackageExtension
        from shutil import rmtree as rmext_tree

        logger.debug('Removing extension %s', name)

        ext = WheelExtension.get_extension(name)

        if not bool(ext):
            msg = 'Extension not found "{0}"'.format(name)
            logger.debug(msg)
            raise ExtensionRemoveError(msg)

        if isinstance(ext, PackageExtension):
            msg = 'Removing a pre-installed extension is prohibited.'
            raise ExtensionRemoveError(msg)

        # Do not remove if installed via azdatadev, could be actual src code
        if ext.edit_mode:
            msg = 'Removing extensions in editable mode (i.e. setuptools ' \
                  '"develop mode") from a local project path is prohibited.'
            logger.debug(msg)
            raise ExtensionRemoveError(msg)

        rmext_tree(
            ext.path,
            onerror=lambda func, path, exc_info: logger.warning(
                'Extension remove error %s: %s - %s', func, path, exc_info)
        )

    @staticmethod
    def list():
        from azdata.cli.core.extensions import BaseExtension
        exts = []
        for ext in BaseExtension.get_extensions():
            exts.append(ext.to_dict())

        return exts
