# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from azdata.cli.core.exceptions import CliError
from azdata.cli.commands.extension import _


def extension_add(client, source, index_url=None, pip_proxy=None,
                  pip_extra_index_urls=None, yes=None):  # pylint: disable=W0613
    """
    ```
    azdata extension add --source <source>
    ```

    Add extension by `source` which can be a file-path to a local Python wheel
    or URL to a remote wheel.

    For more information view: `azdata extension add --help`
    """

    from azdata.cli.core.configuration import Configuration
    from azdata.cli.commands.extension.exceptions import (
        ExtensionInstallError,
        ExtensionExistsError
    )
    from six.moves.urllib.parse import urlparse
    from urllib.error import HTTPError
    from urllib.error import URLError

    import os

    def install(client, source, index_url, pip_proxy, pip_extra_index_urls):
        return client.install(
            source=source,
            index_url=index_url,
            pip_proxy=pip_proxy,
            pip_extra_index_urls=pip_extra_index_urls
        )

    url_tokens = urlparse(source)

    try:
        # -- is source a URL --
        if url_tokens.scheme in ['http', 'https']:
            source = client.utils.download(
                url_tokens.geturl(),
                os.path.basename(url_tokens.path)
            )

        # -- install source wheel --
        ext_name, _ext_path, _target = client.terminal.progress_indicator.\
            message(_('Installing')).\
            worker(install, {
                'client': client,
                'source': source,
                'index_url': index_url,
                'pip_proxy': pip_proxy,
                'pip_extra_index_urls': pip_extra_index_urls,
            }).start()

        client.stdout(_('Extension "{0}" successfully installed. To remove '
                        'this extension:\n{1} extension remove --name {0}').
                      format(ext_name, Configuration().CLI_NAME))

    except (ValueError, ExtensionInstallError, ExtensionExistsError) as ex:
        raise CliError(ex)
    except (URLError, HTTPError):
        raise CliError('Extension failed to download, please verify that a '
                       'valid extension wheel is available at:\n'
                       '[--source, -s] {}'.format(source))


def extension_remove(client, name, yes=None):  # pylint: disable=W0613
    """
    ```
    azdata extension remove --name <name>
    ```

    Remove installed extension by `name`.

    For more information view: `azdata extension remove --help`
    """
    from azdata.cli.commands.extension.exceptions import ExtensionRemoveError

    try:
        client.remove(name)
        client.stdout(_('Extension "{0}" successfully removed.').format(name))
    except ExtensionRemoveError as ex:
        raise CliError(ex)


def extension_list(client):
    """
    ```
    azdata extension list
    ```

    List all installed extensions.

    For more information view: `azdata extension list --help`
    """
    return client.list()
