from kubernetes import client as k8sClient
from kubernetes.client.rest import ApiException as K8sApiException
from azdata.cli.core.exceptions import http_status_codes
from azdata.cli.core.profile import acquire_token_by_basic
from azdata.cli.core.password import PasswordManager
from azdata.cli.core.configuration import Configuration
from azdata.cli.core.constants import (AZDATA_USERNAME, AZDATA_PASSWORD, DOCKER_USERNAME, DOCKER_PASSWORD)
from azdata.cli.core.clients.controller_client import ControllerClient
from azdata.cli.core.deploy import display, is_active_directory_security_enabled
from urllib3.exceptions import NewConnectionError, MaxRetryError

from azdata.cli.core.logging import get_logger
from azdata.cli.core.util import get_controller_endpoint
import azdata.cli.core.deploy as util
import azdata.cli.core.progress as progress_handler
import azdata.cli.core.kubernetes as kubernetes_util
import azdata.cli.core.upgrade as upgrade_util
import base64
import getpass
import json
import os
import socket
import sys
import threading
import time
import yaml

CONNECTION_RETRY_ATTEMPTS = 12
RETRY_INTERVAL = 5
UPDATE_INTERVAL = (15 * 60) / RETRY_INTERVAL
SCRIPT_PATH = os.path.dirname(os.path.realpath(__file__))
logger = get_logger(__name__)


class ControlDeployClient(object):
    """
    The client for managing creation of a control plane
    """
    def __init__(self, cli_client, namespace=None):
        self.cli_client = cli_client
        self.namespace = namespace
        self.logs_lock = threading.Lock()
        self.offset = 0

    class SecurityConfig:
        """
        Wrapper for parsing the security section of the config
        """
        def __init__(self, security_config):
            self.security_config = security_config
            self.useInternalDomain = False

            # Get if deploying to an external domain from the config
            #
            if self.security_config \
                    and "activeDirectory" in security_config \
                    and "useInternalDomain" in security_config["activeDirectory"]:
                self.useInternalDomain = security_config["activeDirectory"]["useInternalDomain"]

            self.domain_controller_ip_addresses = []
            self.calculate_domain_controller_ip_addresses()

        def calculate_domain_controller_ip_addresses(self):
            """
            Lookup the IP of all domain controllers if using an external domain
            :return:
            """
            if not self.useInternalDomain \
                    and self.security_config \
                    and "activeDirectory" in self.security_config \
                    and "domainControllerFullyQualifiedDns" in self.security_config["activeDirectory"]:
                num_dns_servers = 0

                if "dnsIpAddresses" in self.security_config["activeDirectory"]:
                    num_dns_servers = len(self.security_config["activeDirectory"]["dnsIpAddresses"])

                for i in range(0, len(self.security_config["activeDirectory"]["domainControllerFullyQualifiedDns"])):
                    try:
                        # Lookup the IP of the domain controller
                        #
                        self.domain_controller_ip_addresses.append(socket.gethostbyname(self.security_config["activeDirectory"]["domainControllerFullyQualifiedDns"][i]))
                    except socket.gaierror as e:
                        # Could not find IP of domain controller, so fall back to IP of DNS server
                        #
                        if i < num_dns_servers:
                            self.domain_controller_ip_addresses.append(self.security_config["activeDirectory"]["dnsIpAddresses"][i])
                        else:
                            # Could not fallback on DNS IP, so fail
                            #
                            logger.error(e)
                            raise ValueError(
                                "IP of domain controller FQDN '{0}' in the "
                                "active directory config could not be "
                                "resolved. Make sure that the FQDN resolves "
                                "to an IP via DNS. If the domain controller "
                                "and DNS servers are on the same machine, "
                                "make sure the IP of all specified domain "
                                "controllers are specified in dnsIpAddresses "
                                "in control.json".format(
                                    self.security_config["activeDirectory"]["domainControllerFullyQualifiedDns"][i]
                                ))


    def create_control_plane(
            self,
            config,
            namespace=None,
            status_thread=None,
            no_wait=False,
            sharedController=False
    ):
        """
        Create control plane
        :param config:
        :param namespace:
        :return:
        """
        PasswordManager.check_password_validity("cluster")
        cluster_secrets = dict()

        # Overwrite cluster name if it is specified on the command line.
        #
        if namespace is not None and namespace != "":
            config["metadata"]["name"] = namespace
        else:
            namespace = config["metadata"]["name"]

        self.namespace = namespace

        # Setup Private Docker Registry
        #
        docker_registry = \
            config["spec"]["controlPlane"]["spec"]["docker"]["registry"]

        # If deploying in AD mode, define the domain controller IP addresses
        #
        if "security" in config["spec"]["controlPlane"] \
                and "activeDirectory" in config["spec"]["controlPlane"]["security"]:
            security_config = self.SecurityConfig(config["spec"]["controlPlane"]["security"])
            config["spec"]["controlPlane"]["security"]["activeDirectory"]["domainControllerIpAddresses"] = \
                security_config.domain_controller_ip_addresses

        # Initialize and check storage. Ensure storage classes exist for persistent volumes.
        #
        util.retry(self.initialize_and_check_storage,
                   config,
                   retry_count=CONNECTION_RETRY_ATTEMPTS,
                   retry_delay=RETRY_INTERVAL,
                   retry_method="list k8s storage classes",
                   retry_on_exceptions=(NewConnectionError, MaxRetryError))

        # Create Kubernetes namespace
        #
        self.create_namespace(self.namespace, cluster_label_key='MSSQL_CLUSTER')

        if (DOCKER_USERNAME in os.environ and os.environ.get(DOCKER_USERNAME)) and (
                DOCKER_PASSWORD in os.environ and os.environ.get(DOCKER_PASSWORD)):
            util.retry(kubernetes_util.setup_private_registry, namespace, docker_registry,
                       retry_count=CONNECTION_RETRY_ATTEMPTS,
                       retry_delay=RETRY_INTERVAL, retry_method="set up docker private registry",
                       retry_on_exceptions=(NewConnectionError, MaxRetryError))

        # Create controller
        #
        cluster_secrets["clusterName"] = self.namespace
        self.create_controller(config, cluster_secrets, no_wait=no_wait, sharedController=sharedController)

        if no_wait:
            return

        controller_endpoint = get_controller_endpoint(self.namespace)
        ip_endpoint = get_controller_endpoint(self.namespace, True)

        if controller_endpoint == ip_endpoint:
            endpoint_str = controller_endpoint
        else:
            endpoint_str = controller_endpoint + ', ' + ip_endpoint

        progress_handler.notify_service_endpoint_available("controller", endpoint_str)
        
        # Use ip endpoint for basic auth during deployment
        controller_endpoint = ip_endpoint

        # Get the initial throw away controller client based on basic auth for
        # the initial bootstrapping. Once controller has been stood up we ignore
        cfg = Configuration()
        cfg.username = os.environ[AZDATA_USERNAME]
        cfg.password = os.environ[AZDATA_PASSWORD]
        tmp_controller_client = ControllerClient(controller_endpoint)
        tmp_controller_client.set_authorization_header(
            "Basic",
            cfg.get_basic_auth_token(strip_prefix=True)
        )

        i = 0
        while not tmp_controller_client.is_ready():
            # Log to console once every 5 minutes if controller is not ready
            if i != 0 and i % 60 == 0:
                display("Waiting for control plane to be ready after %d "
                        "minutes." % ((i * RETRY_INTERVAL) / 60))

            if sharedController:
                # Log to console once every 1 minutes to display progress.
                # TODO: Right approach should be output progress via get_controller_logs, but it currently disabled in line #131
                # Once it is solved, we should log progress in FSM
                if i != 0 and i % 12 == 0:
                    progress_handler.notify_control_plane_status(tmp_controller_client, controller_endpoint, config)
                
            time.sleep(RETRY_INTERVAL)
            i = i + 1

        if sharedController:
            # Log *status* to console after tmp_controller_client.is_ready()
            progress_handler.notify_control_plane_status(tmp_controller_client, controller_endpoint, config)

        # Update the controller client ref with the endpoint and with an
        # authenticated access-token such that all future calls are allowed
        # access to protected resources via Bearer auth going forward in this
        # process.
        token = acquire_token_by_basic(
            controller_endpoint,
            username=cfg.username,
            password=cfg.password
        )

        self.cli_client.apis.controller.endpoint = controller_endpoint
        self.cli_client.apis.controller.set_authorization_header(
            'Bearer',
            token.access_token
        )

        if sharedController:
            # Create a record in CRD if applicable otherwise ignore
            # TODO: Remove this and place into new `arc` section
            # Other question, should this be here? Perhaps add it to the
            # `control.client` after this function returns rather than stuffing
            # into this tier.
            self.cli_client.apis.controller.register_instance_by_configuration(
                config,
                self.namespace
            )

        # Serialize the clusterObject back into the config
        #
        if not sharedController:
            # TODO put this logic in BDC deployment only.
            if not status_thread:
                stop = threading.Event()
                status_thread = threading.Thread(
                    target=self.get_controller_logs, args=[stop])

            status_thread.setDaemon(True)
            status_thread.start()

    def create_controller(
            self,
            config,
            secrets,
            node_admin_login='nodeadmin',
            knox_username='root',
            no_wait=False,
            sharedController=False):
        """
        Create controller
        :param config: Control config to be deployed
        :param secrets: Secrets of the control
        :param node_admin_login:
        :param knox_username:
        :param sharedController: TODO
        :return:
        """
        if self.is_controller_pod_running(self.namespace):
            display("Controller already exists.")
            return

        # Check necessary environment variables for controller.
        #
        util.check_environment_variables("controller")

        # Check if passed in passwords pass the complexity requirements.
        #
        PasswordManager.check_password_validity("controller")

        # Create controller database SA password
        #
        controller_sa_password = PasswordManager.generate_random_sql_linux_password()
        secrets["controllerSaPassword"] = controller_sa_password

        # Create admin login password
        #
        node_admin_password = PasswordManager.generate_random_sql_linux_password()

        timestamp = str(int(time.time()))

        # Save base64 values in the cluster secrets object
        #
        if is_active_directory_security_enabled(config["spec"]["controlPlane"]):
            secrets["base64DomainServiceAccountPassword"] = base64.b64encode(
                bytes(os.environ["DOMAIN_SERVICE_ACCOUNT_PASSWORD"], 'utf-8')).decode('utf-8')
            secrets["base64DomainServiceAccountUsername"] = base64.b64encode(
                bytes(os.environ["DOMAIN_SERVICE_ACCOUNT_USERNAME"], 'utf-8')).decode('utf-8')
        else:
            secrets["base64DomainServiceAccountPassword"] = ""
            secrets["base64DomainServiceAccountUsername"] = ""

        secrets["base64ControllerUsername"] = base64.b64encode(
            bytes(os.environ["AZDATA_USERNAME"], 'utf-8')).decode('utf-8')
        secrets["base64ControllerPassword"] = base64.b64encode(
            bytes(os.environ["AZDATA_PASSWORD"], 'utf-8')).decode('utf-8')
        secrets["base64ControllerSaPassword"] = base64.b64encode(
            bytes(controller_sa_password, 'utf-8')).decode('utf-8')
        secrets["base64KnoxUsername"] = base64.b64encode(
            bytes(knox_username, 'utf-8')).decode('utf-8')
        secrets["base64KnoxPassword"] = base64.b64encode(
            bytes(os.environ["AZDATA_PASSWORD"], 'utf-8')).decode('utf-8')
        secrets["base64NodeAdminLogin"] = base64.b64encode(
            bytes(node_admin_login, 'utf-8')).decode('utf-8')
        secrets["base64NodeAdminPassword"] = base64.b64encode(
            bytes(node_admin_password, 'utf-8')).decode('utf-8')
        secrets["base64Timestamp"] = base64.b64encode(
            bytes(timestamp, 'utf-8')).decode('utf-8')
        secrets["base64mssqlSaPassword"] = base64.b64encode(
            bytes(os.environ["AZDATA_PASSWORD"], 'utf-8')).decode('utf-8')

        configs = {
            "controllerReplicaSetConfig": "rs-controller.yaml.tmpl",
            "controllerSvcConfig": "svc-controller.yaml.tmpl",
            "controllerEndpointSvcConfig": "svc-controller-external.yaml.tmpl",
            "namespacedAdminRoleConfig": "role-namespaced-admin.yaml",
            "controllerConfigMapConfig": "cm-controller.json.tmpl",
            "controllerDatabaseStatefulSetConfig": "ss-controldb.yaml.tmpl",
            "controllerDatabaseSvcConfig": "svc-controldb.yaml.tmpl"
        }

        if sharedController:
            configs["clusterAdminRoleConfig"] = "role-cluster-admin.yaml"
            configs["evergreenSqlOperatorConfig"] = "evergreen-sql/deployment/mssql-evergreen-operator.yaml.tmpl"
            configs["evergreenSqlCrd"] = "evergreen-sql/deployment/mssql_evergreen_crd.yaml"
            configs["evergreenSqlOperatorRole"] = "evergreen-sql/deployment/mssql-evergreen-operator-role.yaml"
            configs["evergreenSqlOperatorRoleBinding"] = "evergreen-sql/deployment/mssql-evergreen-operator-rolebinding.yaml.tmpl"
            configs["azureArcRegistrationCrd"] = "arc/deployment/registration_crd_deployment.yaml"

        secret_configs = {
            "domainServiceAccountSecretConfig": "secret-domain-service-account.yaml.tmpl",
            "controllerLoginSecretConfig": "secret-controller-login.yaml.tmpl",
            "controllerSaSecretConfig": "secret-controller-sa.yaml.tmpl",
            "nodeAdminLoginSecretConfig": "secret-node-admin-login.yaml.tmpl",
            "mssqlSaSecretConfig": "secret-mssql-sa.yaml.tmpl"
        }

        if "image" not in config["spec"]["controlPlane"]["spec"]["docker"]:
            config["spec"]["controlPlane"]["spec"]["docker"]["image"] = "mssql-controller"

        # Update deployment templates
        #
        for key, value in configs.items():
            template_config = util.get_config_from_template(os.path.join(SCRIPT_PATH, "../templates", value), config)
            configs[key] = template_config

        # Update secret templates
        #
        for key, value in secret_configs.items():
            template_config = util.get_config_from_template(os.path.join(SCRIPT_PATH, "../templates", value), secrets)
            secret_configs[key] = template_config

        display("Starting cluster deployment.")

        # Create secrets for controller
        #
        for _, value in secret_configs.items():
            util.retry(kubernetes_util.create_secret, self.namespace, value,
                       retry_count=CONNECTION_RETRY_ATTEMPTS,
                       retry_delay=RETRY_INTERVAL, retry_method="create secrets",
                       retry_on_exceptions=(NewConnectionError, MaxRetryError))

        # Create empty agent secrets for controller and controller db
        # Controller will modify them with the right data
        #
        util.retry(kubernetes_util.create_empty_secret, self.namespace, "controldb-agent-secret",
                   retry_count=CONNECTION_RETRY_ATTEMPTS,
                   retry_delay=RETRY_INTERVAL, retry_method="create agent secret for controller db",
                   retry_on_exceptions=(NewConnectionError, MaxRetryError))

        util.retry(kubernetes_util.create_empty_secret, self.namespace, "control-agent-secret",
                   retry_count=CONNECTION_RETRY_ATTEMPTS,
                   retry_delay=RETRY_INTERVAL, retry_method="create agent secret for controller",
                   retry_on_exceptions=(NewConnectionError, MaxRetryError))

        # Create controller service account, namespaced role and role binding
        #
        try:
            if not util.retry(kubernetes_util.service_account_exists, self.namespace, "sa-mssql-controller",
                   retry_count=CONNECTION_RETRY_ATTEMPTS,
                   retry_delay=RETRY_INTERVAL, retry_method="check if service account exists",
                   retry_on_exceptions=(NewConnectionError, MaxRetryError)):

                   body = k8sClient.V1ServiceAccount()
                   body.metadata = k8sClient.V1ObjectMeta(name="sa-mssql-controller",
                                                          labels={'MSSQL_CLUSTER': self.namespace})
                   k8sClient.CoreV1Api().create_namespaced_service_account(namespace=self.namespace, body=body)

            if not util.retry(kubernetes_util.namespaced_role_exists, self.namespace, "namespaced-admin",
                   retry_count=CONNECTION_RETRY_ATTEMPTS,
                   retry_delay=RETRY_INTERVAL, retry_method="check if role exists",
                   retry_on_exceptions=(NewConnectionError, MaxRetryError)):

                   body = yaml.full_load(configs["namespacedAdminRoleConfig"])
                   k8sClient.RbacAuthorizationV1Api().create_namespaced_role(namespace=self.namespace, body=body)

            role_binding_name = "rb-" + self.namespace + "-controller"
            if not util.retry(kubernetes_util.namespaced_role_binding_exists, self.namespace, role_binding_name,
                   retry_count=CONNECTION_RETRY_ATTEMPTS,
                   retry_delay=RETRY_INTERVAL, retry_method="check if role binding exists",
                   retry_on_exceptions=(NewConnectionError, MaxRetryError)):

                   body = k8sClient.V1RoleBinding(
                       metadata=k8sClient.V1ObjectMeta(name=role_binding_name,
                                                       namespace=self.namespace,
                                                       labels={'MSSQL_CLUSTER': self.namespace}),
                       subjects=[
                            k8sClient.V1Subject(kind="ServiceAccount", name="sa-mssql-controller",
                                                namespace=self.namespace)],
                       role_ref=k8sClient.V1RoleRef(name="namespaced-admin", kind="Role",
                                                    api_group="rbac.authorization.k8s.io"))
                   k8sClient.RbacAuthorizationV1Api().create_namespaced_role_binding(namespace=self.namespace, body=body)
            
            if sharedController:
                # Grant access to all namespaces
                body = yaml.full_load(configs["clusterAdminRoleConfig"])
                role = k8sClient.RbacAuthorizationV1Api().replace_cluster_role(
                    name=body["metadata"]["name"], body=body)
                
                # Assuming cluster role is already created during azdata control create
                # Create role binding
                roleBinding = self._create_role_binding_object(self.namespace, role, "crb-" + self.namespace + "-controller")
                try:
                    k8sClient.RbacAuthorizationV1Api().create_cluster_role_binding(body=roleBinding)
                except K8sApiException as e:
                    if e.status != http_status_codes.conflict: raise
        except K8sApiException as e:
            logger.error(e.body)
            raise

        # Create cluster role for monitoring
        #
        self.create_cluster_role_for_monitoring(config, self.namespace)

        if sharedController:
            # Create Custom Resources Definition operator.
            try:
                body = yaml.full_load(configs["evergreenSqlOperatorConfig"])
                k8sClient.AppsV1Api().create_namespaced_deployment(body=body, namespace=self.namespace)

                body = yaml.full_load(configs["evergreenSqlOperatorRole"])
                k8sClient.RbacAuthorizationV1Api().create_cluster_role(body=body)

                body = yaml.full_load(configs["evergreenSqlOperatorRoleBinding"])
                k8sClient.RbacAuthorizationV1Api().create_cluster_role_binding(body=body)
            except Exception as e:
                logger.error(e)
                raise

            # Don't raise when CRD creation failed due to k8s version.
            try:
                body = yaml.full_load(configs["evergreenSqlCrd"])
                k8sClient.ApiextensionsV1beta1Api().create_custom_resource_definition(body=body)
            except ValueError as e:
                logger.info(e)

            # Don't raise when CRD creation failed due to k8s version.
            try:
                body = yaml.full_load(configs["azureArcRegistrationCrd"])
                k8sClient.ApiextensionsV1beta1Api().create_custom_resource_definition(body=body)
            except ValueError as e:
                logger.info(e)

        # Create controller pod and service
        #
        try:
            body = json.loads(configs["controllerConfigMapConfig"])

            # Add control plane config to the config map
            #
            body["data"]["mssql-controller.conf"] = json.dumps(config["spec"]["controlPlane"])
            body["metadata"]["labels"] = {'MSSQL_CLUSTER': self.namespace}

            node_label = None
            cluster_label = None

            if "clusterLabel" in config["spec"]["controlPlane"]["spec"]:
                cluster_label = config["spec"]["controlPlane"]["spec"]["clusterLabel"]

            if "nodeLabel" in config["spec"]["controlPlane"]["spec"]:
                node_label = config["spec"]["controlPlane"]["spec"]["nodeLabel"]

            util.add_node_selector_to_controller_config_map(body, cluster_label, node_label)

            # Create controller ConfigMap
            #
            k8sClient.CoreV1Api().create_namespaced_config_map(namespace=self.namespace, body=body)

            # Create Cluster ConfigMap
            #
            data = {}
            kubernetes_util.create_config_map(cluster_name=self.namespace, config_map_name="cluster-configmap",
                                              data=data)

            # Create PVCs if persistent storage is defined.
            #
            if "storage" in config["spec"]["controlPlane"]["spec"]:
                control_pods = ["controller", "controldb"]

                # Create PVs for logs and data for the controller and controller db pods.
                #
                for pod in control_pods:
                    pvc_config = util.get_config_from_template(
                        os.path.join(SCRIPT_PATH, "../templates", "pvc-logs-control.yaml.tmpl"), config)
                    body = yaml.full_load(pvc_config)
                    body["metadata"]["name"] = "logs-%s" % pod
                    body["metadata"]["labels"] = {'MSSQL_CLUSTER': self.namespace}
                    k8sClient.CoreV1Api().create_namespaced_persistent_volume_claim(namespace=self.namespace,
                                                                                    body=body)

                    pvc_config = util.get_config_from_template(
                        os.path.join(SCRIPT_PATH, "../templates", "pvc-data-control.yaml.tmpl"), config)
                    body = yaml.full_load(pvc_config)
                    body["metadata"]["name"] = "data-%s" % pod
                    body["metadata"]["labels"] = {'MSSQL_CLUSTER': self.namespace}
                    k8sClient.CoreV1Api().create_namespaced_persistent_volume_claim(namespace=self.namespace,
                                                                                    body=body)

            # Deploy controller db stateful set
            #
            body = yaml.full_load(configs["controllerDatabaseStatefulSetConfig"])
            util.add_node_selector_to_controller_config(body, cluster_label, node_label)

            k8sClient.AppsV1Api().create_namespaced_stateful_set(namespace=self.namespace, body=body)

            # Deploy controller replica set
            #
            body = yaml.full_load(configs["controllerReplicaSetConfig"])
            util.add_node_selector_to_controller_config(body, cluster_label, node_label)

            k8sClient.AppsV1Api().create_namespaced_replica_set(namespace=self.namespace, body=body)

            # Create controller service
            #
            body = yaml.full_load(configs["controllerSvcConfig"])
            k8sClient.CoreV1Api().create_namespaced_service(namespace=self.namespace, body=body)

            # Create controller db service
            #
            body = yaml.full_load(configs["controllerDatabaseSvcConfig"])
            k8sClient.CoreV1Api().create_namespaced_service(namespace=self.namespace, body=body)

            # Create controller endpoint service
            #
            body = yaml.full_load(configs["controllerEndpointSvcConfig"])
            k8sClient.CoreV1Api().create_namespaced_service(namespace=self.namespace, body=body)
        except K8sApiException as e:
            logger.error(e.body)
            raise

        wait_time = 0

        while not no_wait and not self.is_controller_pod_running(self.namespace):
            time.sleep(RETRY_INTERVAL)
            wait_time = wait_time + RETRY_INTERVAL

            if wait_time % 30 == 0:
                display("Waiting for cluster controller to start.")


    def create_cluster_role_for_monitoring(self, config, namespace):
        """
        Create a cluster role for monitoring
        :param config:
        :param namespace:
        :return:
        """

        # Create cluster role with read permission on the pod/node resources
        #
        if "security" not in config["spec"]["controlPlane"] \
            or "allowPodMetricsCollection" not in config["spec"]["controlPlane"]["security"] \
            or config["spec"]["controlPlane"]["security"]["allowPodMetricsCollection"]:

            service_account_name = "sa-mssql-metricsdc-reader"
            cluster_role_name = namespace + ":cr-mssql-metricsdc-reader"
            cluster_role_binding_name = namespace + ":crb-mssql-metricsdc-reader"

            try:
                body = yaml.full_load(util.get_config_from_template(
                    os.path.join(SCRIPT_PATH, "../templates", "clusterrole-metricsdc-reader.yaml"), cluster_role_name))

                if not util.retry(
                    kubernetes_util.cluster_role_exists, cluster_role_name,
                    retry_count=CONNECTION_RETRY_ATTEMPTS,
                    retry_delay=RETRY_INTERVAL, retry_method="check if cluster role exists",
                    retry_on_exceptions=(NewConnectionError, MaxRetryError)):
                    k8sClient.RbacAuthorizationV1Api().create_cluster_role(body=body)

                body = k8sClient.V1ClusterRoleBinding(
                    metadata=k8sClient.V1ObjectMeta(
                        name=cluster_role_binding_name),
                    subjects=[ k8sClient.V1Subject(
                        kind="ServiceAccount", 
                        name=service_account_name,
                        namespace=namespace)],
                    role_ref=k8sClient.V1RoleRef(
                        kind="ClusterRole",
                        name=cluster_role_name, 
                        api_group="rbac.authorization.k8s.io"))

                if not util.retry(kubernetes_util.cluster_role_binding_exists, cluster_role_binding_name,
                    retry_count=CONNECTION_RETRY_ATTEMPTS,
                    retry_delay=RETRY_INTERVAL, retry_method="check if cluster role binding exists",
                    retry_on_exceptions=(NewConnectionError, MaxRetryError)):
                    k8sClient.RbacAuthorizationV1Api().create_cluster_role_binding(body=body)

            except K8sApiException as e:
                # Telegraf requires the cluster wide role for the pod collection. The azdata might not have sufficient permissions to create
                # the ClusterRole and ClusterRoleBinding for telegraf. If so, only print the warning message and keep on deployment. The pod
                # metrics will not be collected, but can be resumed automatically if the cluster role gets created at anytime. 
                #
                logger.warning(e.body)
                logger.warning("Azdata does not have sufficient permissions to create '%s' ClusterRole and '%s' ClusterRoleBinding. Please ask your "
                    "cluster administrator to manually create them for resuming the pod metrics monitoring. More details: https://aka.ms/sql-bdc-rbac",
                    cluster_role_name, cluster_role_binding_name)
                pass


    def upgrade_control_plane(self, cluster_name, cluster_patch):
        """
        Upgrade cluster
        :param cluster_name:
        :param target_version:
        :param target_repository:
        :return:
        """
        self.cluster_name = cluster_name

        kubernetes_util.validate_namespace(self.cluster_name)

        # Make sure the namespace exists
        #
        if not util.retry(kubernetes_util.namespace_exists, self.cluster_name, retry_count=CONNECTION_RETRY_ATTEMPTS,
                          retry_delay=RETRY_INTERVAL,
                          retry_method="check if namespace exists",
                          retry_on_exceptions=(NewConnectionError, MaxRetryError)):
            raise Exception("Upgrade cluster failed since the namespace %s does not exist." % self.cluster_name)

        if not self.is_controller_pod_running(self.cluster_name):
            raise Exception(
                "Upgrade cluster failed since the controller is not deployed into namespace %s." % self.cluster_name)

        display("Upgrading Control Plane.")

        # Upgrade the control plane
        #
        self.cli_client.apis.controller.upgrade_control_plane(cluster_patch)

        # All control upgrade monitoring will be replaced by calls to the monitoring pod. Until then
        # we will manually check the upgrade config to see the status of the controller upgrade, and the
        # control plane state machine for the rest of the upgrade.
        #
        i = 0
        while upgrade_util.is_controller_upgrading(self.cluster_name):
            if i !=0 and i % UPDATE_INTERVAL == 0:
                logger.warning("Controller is still upgrading "
                               "after %d minutes." % (util.minutes_elapsed(i, RETRY_INTERVAL)))

            time.sleep(RETRY_INTERVAL)
            i = i + 1

        if upgrade_util.check_controller_upgrade_failed(self.cluster_name):
            raise Exception("Control plane upgrade failed. Failed to upgrade controller.")

        # The controller and controller db have been upgraded. Now we need to wait for the rest service
        # to be reenabled
        #
        i = 0
        while not self.cli_client.apis.controller.is_ready():
            time.sleep(RETRY_INTERVAL)
            i = i + 1

        control_plane_state = self.cli_client.apis.controller.get_control_status(control_plane_state=True)

        # It can take a second or two for the state machine to move back into the upgrading state.
        # If the upgrade is still old, retry. We don't pause an upgrade until an image has been
        # updating for at least ten minutes, so we shouldn't get stuck in this loop.
        #
        while control_plane_state == "rollbackcomplete":
            time.sleep(RETRY_INTERVAL)

            control_plane_state = self.cli_client.apis.controller.get_control_status(control_plane_state=True)

        stop = threading.Event()
        status_thread = threading.Thread(
            target=self.get_controller_logs, args=[stop, False])
        status_thread.setDaemon(True)
        status_thread.start()

        i = 0
        while kubernetes_util.is_control_plane_upgrading(self.cluster_name):

            if control_plane_state == "error":
                stop.set()
                status_thread.join()
                progress_handler.notify_control_plane_upgrade_error(self.cluster_name, control_plane_state)
                raise Exception("Control plane upgrade failed. Cluster State: %s." % control_plane_state)
            elif control_plane_state == "rollbackcomplete":
                progress_handler.notify_control_plane_upgrade_paused()
                return

            if i !=0 and i % UPDATE_INTERVAL == 0:
                logger.warning("Control plane is not ready after %d minutes." % util.minutes_elapsed(i, RETRY_INTERVAL))

            if control_plane_state is not None:
                logger.debug("Waiting for Control plane to be in Ready state. Current state : " + control_plane_state)

            time.sleep(RETRY_INTERVAL)
            control_plane_state = self.cli_client.apis.controller.get_control_status(control_plane_state=True)
            i = i + 1

        progress_handler.notify_control_plane_upgrade_success()

    def create_namespace(self, namespace, cluster_label_key=None):
        """
        Create kubernetes namespace
        :param namespace: name of the namespace
        :param cluster_label_key: cluster label key/name
        :return:
        """
        kubernetes_util.validate_namespace(namespace)

        # Create namespaces for the new cluster
        #
        if util.retry(kubernetes_util.namespace_exists, namespace, retry_count=CONNECTION_RETRY_ATTEMPTS,
                      retry_delay=RETRY_INTERVAL,
                      retry_method="check if namespace exists",
                      retry_on_exceptions=(NewConnectionError, MaxRetryError)):
            label_selector = None

            if cluster_label_key:
                label_selector = cluster_label_key + "=" + namespace

            if util.retry(kubernetes_util.namespace_is_empty, namespace, label_selector,
                          retry_count=CONNECTION_RETRY_ATTEMPTS, retry_delay=RETRY_INTERVAL,
                          retry_method="check if namespace is empty",
                          retry_on_exceptions=(NewConnectionError, MaxRetryError)):
                if cluster_label_key:
                    namespace_response = k8sClient.CoreV1Api().read_namespace(namespace)
                    if namespace_response.metadata.labels is None or cluster_label_key not in namespace_response.metadata.labels:
                        display("NOTE: Namespace \"%s\" is already created and will not be labeled with \"%s\" Kubernetes label." % (namespace, cluster_label_key))
                        display("This is an informational message only, no user action is required.")
                        display("")
            else:
                raise Exception(
                    "Cluster creation not initiated because the existing namespace %s "
                    "is not empty. Run \"kubectl get all -n %s \""
                    " to see the existing objects in the namespace" %
                    (namespace, namespace))
        else:
            labels = {cluster_label_key: namespace} if cluster_label_key else None
            util.retry(kubernetes_util.create_namespace, namespace, labels,
                       retry_count=CONNECTION_RETRY_ATTEMPTS,
                       retry_delay=RETRY_INTERVAL, retry_method="create namespace",
                       retry_on_exceptions=(NewConnectionError, MaxRetryError))

    @staticmethod
    def is_controller_pod_running(namespace):
        """
        Check if controller is running.
        :return:
        """
        try:
            controller_service = k8sClient.CoreV1Api().read_namespaced_service(name="controller-svc-external",
                                                                               namespace=namespace)
            if controller_service.spec.type == 'LoadBalancer':
                if controller_service.status.load_balancer.ingress is None:
                    logger.debug(
                        "Waiting for public IP to be assigned for the controller service, please check the kubernetes "
                        "events for more details.")
                    return False

            # check if replicas are available for controller
            #
            rs_output = k8sClient.AppsV1Api().read_namespaced_replica_set(name="control", namespace=namespace)
            if rs_output.status.available_replicas != rs_output.status.replicas:
                logger.debug(
                    "Controller pod(s) are not in running state yet, please check the kubernetes events for more "
                    "details.")
                return False

            return True

        except K8sApiException as e:
            logger.debug(e.body)
            return False

    def get_controller_logs(self, stop, execute_once=False):
        """
        Read Controller logs
        :param stop: threading stop event, when triggered it will stop the infinite loop
        :param execute_once:
        :return:
        """
        try:
            while not stop.is_set():
                try:
                    self.logs_lock.acquire()
                    response = self.cli_client.apis.controller.\
                        get_controller_logs(self.offset)
                    progress_handler.log_progress(response)
                    if not response:
                        pass
                    elif not response.endswith('\r\n'):
                        self.offset += len(response.rsplit('\r\n', 1)[0])
                        continue
                    else:
                        self.offset += len(response)
                finally:
                    self.logs_lock.release()
                if execute_once:
                    return
                time.sleep(RETRY_INTERVAL)
        except Exception as e:
            display(e)

    def set_password(self, service_name, knox_username):
        """
        Updates password for the given user
        :param service_name:
        :param knox_username:
        :return:
        """

        # Check necessary environment variables for controller.
        #
        util.check_environment_variables("password")

        # Validate cluster name
        #
        if self.namespace.lower() == "default":
            raise Exception("Cluster name can not be \"default\".")

        # Validate that the controller is running and get its endpoint
        #
        if not self.is_controller_pod_running(self.namespace):
            raise Exception("Controller is not running.")

        # Read username and the new password
        #
        if service_name == "controller":
            password_var = "CONTROLLER_NEW_PASSWORD"
        else:
            raise Exception(
                "Unsupported service for set-password: '%s'. Valid services are {controller}." % service_name)

        if password_var in os.environ:
            new_password = os.environ[password_var]
        else:
            new_password = getpass.getpass("Enter new password: ")
            if new_password != getpass.getpass("Re-enter new password: "):
                raise Exception("The passwords do not match.")

        # Run password complexity check
        #
        if not PasswordManager.validate_password_and_print_if_error(new_password):
            logger.warning("Password complexity check failed. Exiting...")
            sys.exit(1)

        # Issue password update
        #
        if service_name == "controller":
            self.set_password_through_controller(service_name, new_password, knox_username)
        else:
            raise Exception(
                "Unsupported service for set-password: '%s'. Valid services are: {controller}." % service_name)

    @staticmethod
    def initialize_and_check_storage(config):
        """
        Check if storage class exists.
        :param config:
        :return:
        """

        control_plane = config["spec"]["controlPlane"]
        if "storage" in control_plane["spec"]:
            storage = control_plane["spec"]["storage"]

            if "data" not in storage:
                config["spec"]["controlPlane"]["spec"]["storage"]["data"] = {}
                config["spec"]["controlPlane"]["spec"]["storage"]["data"]["className"] = storage[
                    "className"]
                config["spec"]["controlPlane"]["spec"]["storage"]["data"]["accessMode"] = storage[
                    "accessMode"]
                config["spec"]["controlPlane"]["spec"]["storage"]["data"]["size"] = storage["size"]

            if "logs" not in storage:
                config["spec"]["controlPlane"]["spec"]["storage"]["logs"] = {}
                config["spec"]["controlPlane"]["spec"]["storage"]["logs"]["className"] = storage[
                    "className"]
                config["spec"]["controlPlane"]["spec"]["storage"]["logs"]["accessMode"] = storage[
                    "accessMode"]
                config["spec"]["controlPlane"]["spec"]["storage"]["logs"]["size"] = storage["size"]

            storage = config["spec"]["controlPlane"]["spec"]["storage"]

    @staticmethod
    def _create_role_binding_object(namespace, role, roleBindingName):
        return k8sClient.V1RoleBinding(
            metadata=k8sClient.V1ObjectMeta(
                name=roleBindingName,
                namespace=namespace,
                labels={'MSSQL_CLUSTER': namespace}),
            subjects=[
                k8sClient.V1Subject(
                    kind="ServiceAccount", name="sa-mssql-controller", namespace=namespace)],
            role_ref=k8sClient.V1RoleRef(
                name=role.metadata.name, kind=role.kind,
                api_group="rbac.authorization.k8s.io"))
