# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------
from kubernetes import client as k8sClient
from azdata.cli.core.exceptions import (KubernetesError)
from kubernetes.client.rest import ApiException as K8sApiException
from azdata.cli.core.exceptions import http_status_codes
from azdata.cli.core.logging import get_logger

logger = get_logger(__name__)


class KubernetesClient(object):
    """
    Client for managing Kubernetes Resources, to be used by Arc partner command modules.
    """

    """
    Custom Resource Definition specification key.
    """
    CUSTOM_RESOURCE_DEFINITION = 'CustomResourceDefinition'

    @staticmethod
    def get_service(ns, service_name):
        """
        Get the specified service in the given namespace.
        :param ns:
        :return:
        """
        try:
            return k8sClient.CoreV1Api().read_namespaced_service(service_name, ns)
        except K8sApiException as e:
            logger.debug(e.body)
            raise

    @staticmethod
    def service_exists(ns, service_name):
        """
        Get the specified service in the given namespace.
        :param ns:
        :return:
        """
        try:
            service = k8sClient.CoreV1Api().read_namespaced_service(service_name, ns)
            return service_name == service.metadata.name
        except K8sApiException as e:
            logger.debug(e.body)
            return False

    @staticmethod
    def service_ready(ns, service_name):
        """
        Check if the specified service is ready
        :param ns:
        :param service_name:
        :return:
        """
        try:
            service = KubernetesClient.get_service(ns, service_name)

            if service.spec.type == 'LoadBalancer':
                # Make sure that the load balancer has at least one host listed
                #
                ingress = service.status.load_balancer.ingress
                if ingress is not None and len(ingress) > 0:
                    for svc in ingress:
                        if svc.ip or svc.hostname:
                            return True
            elif service.spec.type == 'NodePort':
                # No additional checks are required for NodePorts
                #
                return True

            return False
        except K8sApiException as e:
            return False

    @staticmethod
    def get_service_endpoint(ns, service, force_ip=False, app='controller'):
        """
        Returns the endpoint of the service based on whether it is LoadBalancer or NodePort
        :param ns:
        :param service:
        :param force_ip:
        :param app:
        :return:
        """
        host = None
        port = None
        if service.spec.type == 'LoadBalancer':
            ingress = service.status.load_balancer.ingress
            if ingress is not None and len(ingress) > 0:
                for svc in ingress:
                    host = svc.ip or svc.hostname
                    if host:
                        port = str(service.spec.ports[0].port)
                        break
        elif service.spec.type == 'NodePort':
            # Get the IP of any pod in the controller replica set for NodePort
            # service
            controller_pods = KubernetesClient.list_pods(ns, 'app={}'.format(app)).items
            if controller_pods and len(controller_pods) > 0:
                host = controller_pods[0].status.host_ip
                port = str(service.spec.ports[0].node_port)
        else:
            raise Exception('Service type {0} for service '
                            '{1} is not supported.'.format(service.spec.type, service.metadata.name))

        if service.metadata.labels.get('dnsName', None) and not force_ip:
            host = service.metadata.labels['dnsName']

        if not host:
            raise ValueError("Failed to retrieve service endpoint. Host not found.")
        elif not port:
            raise ValueError("Failed to retrieve service endpoint. Port not found.")

        return 'https://' + host + ':' + port

    @staticmethod
    def list_pods(ns, label_selector=None):
        """
        List the pods in the given namespace
        :param ns: The namespace
        :param label_selector: The label to select the pods by.
        :return:
        """
        try:
            return k8sClient.CoreV1Api().list_namespaced_pod(ns, label_selector=label_selector)
        except K8sApiException as e:
            logger.error(e.body)
            raise

    @staticmethod
    def pod_is_running(ns, app_label):
        """
        Returns true if the pod is a Running state
        :param ns:
        :param app:
        :return:
        """
        selector = 'app=%s' % app_label
        pods = KubernetesClient.list_pods(ns, selector).items
        return pods and len(pods) > 0 and all(pod.status.phase.lower() == "running" and pod.status.host_ip for pod in pods)


    @staticmethod
    def create_secret(ns, config, ignore_conflict=False):
        """
        Creates a secret in Kubernetes
        :param ignore_conflict:
        :param ns:
        :param config:
        :return:
        """
        try:
            k8sClient.CoreV1Api().create_namespaced_secret(namespace=ns, body=config)
        except K8sApiException as e:
            if not (ignore_conflict and e.status == http_status_codes.conflict):
                raise

    @staticmethod
    def secret_exists(ns, secret_name):
        """
        Returns true if secret exists in the namespace
        :param ns:
        :param secret_name:
        :return:
        """
        try:
            secret = k8sClient.CoreV1Api().read_namespaced_secret(secret_name, ns)
            return secret_name == secret.metadata.name
        except K8sApiException as e:
            logger.debug(e.body)
            return False

    @staticmethod
    def storage_class_exists(class_name):
        """
        Returns true if the storage class exists
        :param class_name:
        :return:
        """
        try:
            storage_class = k8sClient.StorageV1Api().read_storage_class(class_name)
            return storage_class and storage_class.metadata.name == class_name
        except K8sApiException as e:
            logger.debug(e.body)
            return False

    @staticmethod
    def patch_secret(ns, secret_name, body):
        """
        Patches a secret in Kubernetes
        :param ns: The secret's namespace
        :param secret_name: The secret's name
        :param body: A dictionary of fields to patch
        :return: The patched secret
        """
        try:
            return k8sClient.CoreV1Api().patch_namespaced_secret(namespace=ns, name=secret_name, body=body)
        except K8sApiException as e:
            logger.debug(e.body)
            raise

    @staticmethod
    def create_replica_set(ns, spec):
        """
        Creates a replica set in the given namespace
        :param ns:
        :param spec:
        :return:
        """
        try:
            k8sClient.AppsV1Api().create_namespaced_replica_set(ns, spec)
        except K8sApiException as e:
            logger.debug(e.body)
            raise e

    @staticmethod
    def replica_set_exists(ns, replica_set_name):
        """
        Returns true if replica set exists in the namespace
        :param ns:
        :param replica_set_name:
        :return:
        """
        try:
            replica_set = k8sClient.AppsV1Api().read_namespaced_replica_set(replica_set_name, ns)
            return replica_set_name == replica_set.metadata.name
        except K8sApiException as e:
            logger.debug(e.body)
            return False

    @staticmethod
    def create_namespaced_service_account(ns, spec):
        """
        Creates namespaced service account
        :param ns:
        :param spec:
        :return:
        """
        try:
            k8sClient.CoreV1Api().create_namespaced_service_account(ns, spec)
        except K8sApiException as e:
            logger.debug(e.body)
            raise e

    @staticmethod
    def service_account_exists(ns, service_account_name):
        """
        Returns true if service account exists in the namespace
        :param ns:
        :param service_account_name:
        :return:
        """
        try:
            service_account = k8sClient.CoreV1Api().read_namespaced_service_account(service_account_name, ns)
            return service_account_name == service_account.metadata.name
        except K8sApiException as e:
            logger.debug(e.body)
            return False

    @staticmethod
    def create_namespaced_role(ns, spec):
        """
        Creates namespaced role binding
        :param ns:
        :param spec:
        :return:
        """
        try:
            k8sClient.RbacAuthorizationV1Api().create_namespaced_role(ns, spec)
        except K8sApiException as e:
            logger.debug(e.body)
            raise e

    @staticmethod
    def namespaced_role_exists(ns, role_name):
        """
        Returns true if role exists in the namespace
        :param ns:
        :param role_name:
        :return:
        """
        try:
            role = k8sClient.RbacAuthorizationV1Api().read_namespaced_role(role_name, ns)
            return role_name == role.metadata.name
        except K8sApiException as e:
            logger.debug(e.body)
            return False

    @staticmethod
    def create_namespaced_role_binding(ns, spec):
        """
        Creates namespaced role binding
        :param ns:
        :param spec:
        :return:
        """
        try:
            k8sClient.RbacAuthorizationV1Api().create_namespaced_role_binding(ns, spec)
        except K8sApiException as e:
            logger.debug(e.body)
            raise e

    @staticmethod
    def namespaced_role_binding_exists(ns, role_binding_name):
        """
        Returns true if role binding exists in the namespace
        :param ns:
        :param role_binding_name:
        :return:
        """
        try:
            role_binding = k8sClient.RbacAuthorizationV1Api().read_namespaced_role_binding(role_binding_name,
                                                                                           ns)
            return role_binding_name == role_binding.metadata.name
        except K8sApiException as e:
            logger.debug(e.body)
            return False

    # ------------------------------
    # Custom Resource Methods
    # ------------------------------

    @staticmethod
    def create_custom_resource_definition(crd):
        """
        Create the given crd.
        :param crd: The crd to create.
        :return:
        """
        try:
            api = k8sClient.ApiextensionsV1beta1Api()
            current_crds = [x['spec']['names']['kind'].lower()
                            for x in api.list_custom_resource_definition().to_dict()['items']]
            if crd.kind.lower() not in current_crds:
                return api.create_custom_resource_definition(crd.body)

        except ValueError as e:
            # Kubernetes API bug with prior versions of Kubernetes
            # The CRD does get created but returns an erroneous value error that we can safely ignore.
            # This is fixed in Kubernetes version 1.16
            # https://github.com/kubernetes-client/python/issues/1022
            # We can remove this condition when we upgrade to version 1.16, server-side.
            if 'Invalid value for `conditions`, must not be `None`' in str(e):
                return
            else:
                raise e
        except K8sApiException as e:
            if e.status != http_status_codes.conflict:
                raise
            else:
                raise KubernetesError(e)


    @staticmethod
    def delete_custom_resource_definition(crd):
        """
        Delete the given crd.
        :param crd: The crd to create.
        :return:
        """
        try:
            api = k8sClient.ApiextensionsV1beta1Api()
            current_crds = [x['spec']['names']['kind'].lower()
                            for x in api.list_custom_resource_definition().to_dict()['items']]
            if crd.kind.lower() in current_crds:
                return api.delete_custom_resource_definition(crd.name)

        except ValueError as e:
            # Kubernetes API bug with prior versions of Kubernetes
            # The CRD does get created but returns an erroneous value error that we can safely ignore.
            # This is fixed in Kubernetes version 1.16
            # https://github.com/kubernetes-client/python/issues/1022
            # We can remove this condition when we upgrade to version 1.16, server-side.
            if 'Invalid value for `conditions`, must not be `None`' in str(e):
                return
            else:
                raise e
        except Exception as e:
            raise KubernetesError(e)

    @staticmethod
    def create_namespaced_custom_object(crd, cr, ignore_conflict=False):
        """
        Creates a kubernetes custom resource object with the given crd and specification json
        :param ignore_conflict: ignore conflict
        :param crd: The definition of the custom resource to be deployed.
        :param cr: The custom resource instance to be deployed.
        :return:
        """
        try:
            api = k8sClient.CustomObjectsApi()
            return api.create_namespaced_custom_object(
                body=cr.body,
                namespace=cr.namespace,
                plural=crd.plural,
                group=cr.group,
                version=cr.version
            )

        except K8sApiException as e:
            if ignore_conflict and e.status == http_status_codes.conflict:
                return None
            else:
                raise KubernetesError(e)

    @staticmethod
    def patch_namespaced_custom_object(crd, cr):
        """
        Patches a kubernetes custom resource object with the given crd and specification json
        :param crd: The definition of the custom resource to be deployed.
        :param cr: The custom resource instance to be deployed.
        :return:
        """
        try:
            api = k8sClient.CustomObjectsApi()
            return api.patch_namespaced_custom_object(
                body=cr.body,
                namespace=cr.namespace,
                plural=crd.plural,
                group=cr.group,
                version=cr.version,
                name=cr.name
            )

        except Exception as e:
            raise KubernetesError(e)

    @staticmethod
    def replace_namespaced_custom_object(crd, cr):
        """
        Replaces a kubernetes custom resource object with the given crd and specification json
        :param crd: The definition of the custom resource to be deployed.
        :param cr: The custom resource instance to be deployed.
        :return:
        """
        try:
            api = k8sClient.CustomObjectsApi()
            return api.replace_namespaced_custom_object(
                body=cr.body,
                namespace=cr.namespace,
                plural=crd.plural,
                group=cr.group,
                version=cr.version,
                name=cr.name
            )

        except Exception as e:
            raise KubernetesError(e)

    @staticmethod
    def delete_namespaced_custom_object(crd, name, namespace):
        """
        Deletes a custom resource object
        :param crd: The definition of the custom resource to be deleted.
        :param name: The name of the custom resource to be deleted.
        :param namespace: The namespace from which to delete the custom resource.
        :return:
        """
        try:
            api = k8sClient.CustomObjectsApi()
            return api.delete_namespaced_custom_object(
                group=crd.group,
                version=crd.version,
                plural=crd.plural,
                namespace=namespace,
                name=name,
                body=k8sClient.V1DeleteOptions()
            )

        except Exception as e:
            raise KubernetesError(e)

    @staticmethod
    def get_namespaced_custom_object(name, namespace, crd=None, group=None, version=None, plural=None):
        """
        Gets the custom resource object
        :param crd: The definition of the custom resource.
        :param name: The name of the custom resource.
        :param namespace: The namespace of the custom resource.
        :param group: The API version group.
        :param version: The kubernetes custom resource api version.
        :param plural: The plural name of the custom resource definition.
        :return:
        """
        try:
            api = k8sClient.CustomObjectsApi()

            if crd:
                group = crd.group
                version = crd.version
                plural = crd.plural
            elif not group or not version or not plural:
                raise ValueError("Please specify either a valid CRD or the group, version, and plural.")

            return api.get_namespaced_custom_object(
                group=group,
                version=version,
                name=name,
                namespace=namespace,
                plural=plural,
            )

        except K8sApiException as e:
            raise

    @staticmethod
    def namespaced_custom_object_exists(name, namespace, crd=None, group=None, version=None, plural=None):
        """
        Check if the custom resource object exists
        :param crd: The definition of the custom resource.
        :param name: The name of the custom resource.
        :param namespace: The namespace of the custom resource.
        :param group: The API version group.
        :param version: The kubernetes custom resource api version.
        :param plural: The plural name of the custom resource definition.
        :return:
        """
        try:
            custom_object = KubernetesClient.get_namespaced_custom_object(name, namespace, crd, group, version, plural)
            return name == custom_object["metadata"]["name"]
        except K8sApiException as e:
            if e.status == http_status_codes.not_found:
                return False
            else:
                raise

    @staticmethod
    def list_namespaced_custom_object(namespace, crd=None, group=None, version=None, plural=None):
        """
        Lists the custom resource object
        :param crd: The definition of the custom resource.
        :param namespace: The namespace of the custom resource.
        :param group: The API version group.
        :param version: The kubernetes custom resource api version.
        :param plural: The plural name of the custom resource definition.
        :return:
        """
        try:
            api = k8sClient.CustomObjectsApi()

            if crd:
                group = crd.group
                version = crd.version
                plural = crd.plural
            elif not group or not version or not plural:
                raise ValueError("Please specify either a valid CRD or the group, version, and plural.")

            return api.list_namespaced_custom_object(
                group=group,
                version=version,
                namespace=namespace,
                plural=plural,
            )

        except K8sApiException as e:
            logger.debug(e.body)
            raise e
