# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from azdata.cli.core.extensions import WheelExtension
from azdata.cli.core.logging import get_logger

import knack.commands

logger = get_logger(__name__)

__all__ = ['CliCommandsLoader', 'CliCommandGroup']


class CliCommandsLoader(knack.commands.CLICommandsLoader):
    """CLI command loader, containing the command mappings."""

    def __init__(self, *args, **kwargs):
        self._custom_command = kwargs.get('custom_command_type')
        self._loaders_to_cmd = {}
        super(CliCommandsLoader, self).__init__(
            *args,
            excluded_command_handler_args=['client'],
            **kwargs
        )

    def load_command_table(self, args):
        """Load all CLI commands by grouping."""

        if self.command_table:
            return self.command_table

        self.command_table = CliCommandTable(self, args).\
            load_commands(). \
            load_extensions(). \
            to_dict()

        return self.command_table

    def load_arguments(self, command):
        """Load all CLI arguments for a command."""

        loader = self._loaders_to_cmd.get(command, None)
        if loader and command in self.command_table:
            loader.command_name = command
            self.command_table[command].load_arguments()

            loader.load_arguments(command)

            self.argument_registry.arguments.update(
                loader.argument_registry.arguments)
            self.extra_argument_registry.update(loader.extra_argument_registry)

            super(CliCommandsLoader, self).load_arguments(command)

    def command_loader_factory(self, name, prefix):
        """
        Creates a new subclassed instance of the concrete `CliCommandLoader`
        given the Python module `name` and `prefix`.

        :param name: The name of the Python command module to load.
        :param prefix: The prefix of the Python module namespace to load.
        :return: `CliCommandLoader` instance for a command.
        """
        from importlib import import_module
        return getattr(
            import_module(prefix + name),
            'COMMAND_LOADER_CLS',
            None
        )(cli_ctx=self.cli_ctx)


# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


class CliCommandGroup(knack.commands.CommandGroup):
    def command(self, name, handler_name, **kwargs):
        """
        Register a command into the command table

        :param name: The name of the command
        :type name: str
        :param handler_name: The name of the handler that will be  applied to
               the operations template
        :type handler_name: str
        :param kwargs: Kwargs to apply to the command. Possible values:
               `client_factory`, `arguments_loader`, `description_loader`,
               `description`,`formatter_class`, `table_transformer`,
               `deprecate_info`, `validator`, `confirmation`.
        """

        # Stash custom command output markers by command namespace into the
        # cli's invocation context.
        # Custom outputs found in command `output` can be:
        # output=fn|str=table,json,tsv|azdata.cli.core.layout.BaseLayout
        cmd = ' '.join(handler_name.split('_'))
        invocation = self.command_loader.cli_ctx.invocation
        custom = invocation.data.get('custom', {})
        custom[cmd] = kwargs.get('output')
        invocation.data['custom'] = custom

        super(CliCommandGroup, self).command(name, handler_name, **kwargs)

# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


class CliCommandTable(object):
    """CLI command table abstraction."""
    def __init__(self, loader, args):
        self._loader = loader
        self._args = args

    def to_dict(self):
        return self._loader.command_table

    def load_commands(self):
        """
        Populate all core commands into the loader's command-table graph.
        """
        from importlib import import_module
        import pkgutil
        import traceback

        installed_command_modules = []

        try:
            mods_ns_pkg = import_module('azdata.cli.commands')
            installed_command_modules = [modname for _, modname, _ in
                                         pkgutil.iter_modules(
                                             mods_ns_pkg.__path__)]
        except ImportError:
            pass

        for mod_name in installed_command_modules:
            self._load_command_modules(mod_name)

        return self

    def load_extensions(self):
        """
        Populate all extensions into the loader's command-table graph.
        """
        from azdata.cli.core.output import OutputProducer
        from importlib import (import_module, reload)
        import sys

        for ext in WheelExtension.get_extensions():
            try:
                ext_name = ext.name
                ext_dir = ext.path
                ext_module_name = ext.module_name
                ext_reload = ext.reload

                if not ext.is_eligible():
                    continue

                # Load the <non-pre-installed> package extensions. These are
                # the extensions that were added via `extension add`
                if ext_dir:
                    # Need to reimport and reload should a "pre-installed"
                    # extension be "updated" with a new extension via
                    # `extension add`. This allows us to use the updated/newly
                    # added extension over the one that came "pre-installed".
                    if ext_reload:
                        sys.path.insert(0, ext_dir)
                        logger.debug('Reloading module = %s', ext_module_name)
                        reload(import_module(ext_module_name))
                    else:
                        sys.path.append(ext_dir)

                logger.debug('================================================')
                logger.debug('Appended to system path = %s', ext_dir)
                logger.debug('System Path = %s', sys.path)
                logger.debug('Loading extension = %s\n%s\n', ext_name, ext)
                logger.debug('================================================')

                self._load_command_modules(
                    ext_module_name,
                    prefix='',
                    metadata=ext.metadata
                )

            except ImportError as e:
                logger.warn(e)
                logger.warn('Extension "%s" cannot be loaded.', ext.name)
                pass

        return self

    def _load_command_modules(
            self, mod_name, prefix='azdata.cli.commands.', metadata=None):
        """
        The worker to populates the cli command-table and command-group-table
        with a given Python module name and prefix namespace.
        """
        def _load_table_catalogs():
            command = ''
            for arg in self._args:
                if arg.startswith('-'):
                    break
                command = '{} {}'.format(command, arg).strip()

            command_loader = \
                self._loader.command_loader_factory(mod_name, prefix)
            command_table = command_loader.load_command_table(self._args)

            if command in command_table:
                self._loader._loaders_to_cmd[command] = command_loader

            return command_table, command_loader.command_group_table

        # make copies to merge with updated values
        command_table_cp = self._loader.command_table.copy()
        command_group_table_cp = self._loader.command_group_table.copy()

        command_table, command_group_table = _load_table_catalogs()

        # metadata takes precedence. Enforce metadata `preview` over anything
        # configured in the source code's `CommandGroup` command construction
        # in `command_table.py`. We do this to enforce an precedence override.
        if metadata and metadata.get('isPreview', False):
            for command, command_group in command_group_table.items():
                if isinstance(command_group, dict):
                    continue

                if not command_group.group_kwargs.get('is_preview', False):
                    from knack.preview import PreviewItem
                    command_group.group_kwargs['is_preview'] = True
                    command_group.group_kwargs['preview_info'] = PreviewItem(
                        object_type='command group',
                        target=command,
                        cli_ctx=self._loader.cli_ctx
                    )

        # merge in update
        command_table_cp.update(command_table)
        command_group_table_cp.update(command_group_table)
        self._loader.command_table = command_table_cp
        self._loader.command_group_table = command_group_table_cp
