# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from __future__ import absolute_import

from azdata.cli.core.prompt import (prompt, prompt_pass)
from azdata.cli.core.util import (FileUtil, is_windows)
from azdata.cli.core.configuration import Configuration
from azdata.cli.core.enums import SecurityStrategy
from azdata.cli.core.exceptions import (CliError, http_status_codes)
from azdata.cli.core.logging import get_logger
from azdata.cli.core.exceptions import ApiError
from azdata.cli.core.constants import (AZDATA_USERNAME, AZDATA_PASSWORD)
from knack.util import ensure_dir
from knack.prompting import NoTTYException
from abc import ABCMeta, abstractmethod
from six import add_metaclass

import errno
import json
import os


log = get_logger(__name__)

__all__ = [
    'select_authenticator',
    'acquire_token_by_basic',
    'acquire_token_by_ad',
    'acquire_token_by_challenge',
    'CliProfile'
]


class CliProfileContext(object):
    """
    An authenticated user context, located in the user profile, to be used with
    command API calls
    """

    ENDPOINT = 'endpoint'
    USERNAME = 'username'
    SECURITY_STRATEGY = 'securityStrategy'
    ACCESS_TOKEN = 'accessToken'
    REFRESH_TOKEN = 'refreshToken'
    EXPIRES_IN = 'expiresIn'
    EXPIRES_ON = 'expiresOn'
    TOKEN_TYPE = 'tokenType'
    NAMESPACE = 'namespace'
    CONTEXT_TYPE = 'contextType'

    def __init__(
            self,
            endpoint=None,
            username=None,
            security_strategy=None,
            access_token=None,
            refresh_token=None,
            expires_on=None,
            expires_in=None,
            token_type=None,
            context_type=None,
            context_name=None):

        self._endpoint = endpoint
        self._username = username
        self._security_strategy = security_strategy
        self._access_token = access_token
        self._refresh_token = refresh_token
        self._expires_on = expires_on
        self._expires_in = expires_in
        self._token_type = token_type
        self._namespace = context_name
        self._context_type = context_type

    def serialize(self):
        """
        Serializes object into a dictionary for outputting to json file
        :return: A `dict`.
        """
        context = dict()

        context[self.ENDPOINT] = self.endpoint
        context[self.USERNAME] = self.username
        context[self.SECURITY_STRATEGY] = self.security_strategy.value
        context[self.ACCESS_TOKEN] = self.access_token
        context[self.REFRESH_TOKEN] = self.refresh_token
        context[self.EXPIRES_ON] = self.expires_on
        context[self.EXPIRES_IN] = self.expires_in
        context[self.TOKEN_TYPE] = self.token_type
        context[self.NAMESPACE] = self.namespace
        context[self.CONTEXT_TYPE] = self.context_type

        return context

    @property
    def endpoint(self):
        """
        Gets the authenticated endpoint
        """
        return self._endpoint

    @property
    def username(self):
        """
        Gets the authenticated username
        """
        return self._username

    @property
    def security_strategy(self):
        """
        Gets the current security-strategy from the user profile. Consumers of
        this call can apply can perform enumerated checks for equality
        Example:

        ```
        profile = CliProfile()
        profile.security_strategy == SecurityStrategy.ad
        profile.security_strategy == SecurityStrategy.basic
        ```
        :return: The current `SecurityStrategy` otherwise `None`.
        """
        try:
            return SecurityStrategy(self._security_strategy)
        except Exception as e:
            # suppress any invalid types
            log.debug(e)
            return None

    @property
    def access_token(self):
        """
        Gets the authenticated user's access-token.
        """
        return self._access_token

    @property
    def refresh_token(self):
        """
        Gets the authenticated user's refresh-token
        """
        return self._refresh_token

    @property
    def expires_on(self):
        return self._expires_on

    @property
    def expires_in(self):
        return self._expires_in

    @property
    def token_type(self):
        return self._token_type

    @property
    def namespace(self):
        """
        Gets the context name.
        """
        return self._namespace

    @property
    def context_type(self):
        """
        Gets the type of context i.e. 'bdc'
        """
        return self._context_type


class CliProfile(object):
    def __init__(
            self,
            accept_eula=None,
            endpoint=None,
            username=None,
            security_strategy=None,
            auth_token=None,
            context_type=None):

        self._profile_cache = _GLOBAL_CLI_PROFILE_CACHE

        # If this is a new context with a token, set the active context and
        # add the new context to the current contexts
        if auth_token:
            # New user profiles get persisted when an access-token is provided,
            # here we assert all other necessary values are also provided.
            for arg in [accept_eula, endpoint, username, security_strategy,
                        context_type]:
                if arg is None:
                    raise ValueError('Argument "{}" is required.'.format(arg))

            pc = CliProfileConfiguration

            profile = dict()
            profile[pc.ACCEPT_EULA] = accept_eula

            # Create a new active context from the given parameters
            active_context = CliProfileContext(
                endpoint=endpoint.rstrip('/'),
                username=username,
                security_strategy=SecurityStrategy[security_strategy].value,
                access_token=auth_token.access_token,
                refresh_token=None,  # @TODO: auth_token.refresh_token
                expires_on=auth_token.expires_on,
                expires_in=auth_token.expires_in,
                token_type=auth_token.token_type,
                context_name=auth_token.namespace,
                context_type=context_type
            ).serialize()

            # Set the new active context
            profile[pc.ACTIVE_CONTEXT] = active_context

            # Add the new active context to the contexts dictionary. If a
            # context with this namespace already existed, this will override it
            contexts = self.contexts if self.contexts else dict()
            contexts[auth_token.namespace] = active_context
            profile[pc.CONTEXTS] = contexts

            # -- persist the profile in the cache --
            self._profile_cache.set_all(profile)

        # Otherwise, don't set anything in the cache and just create an empty
        # object that accesses the cache

    @property
    def active_context(self):
        """
        Gets the active context.
        """
        cache = self._profile_cache.get(CliProfileConfiguration.ACTIVE_CONTEXT)
        return CliProfileContext() if not cache else CliProfileContext(
            endpoint=cache.get(CliProfileContext.ENDPOINT),
            username=cache.get(CliProfileContext.USERNAME),
            security_strategy=cache.get(CliProfileContext.SECURITY_STRATEGY),
            access_token=cache.get(CliProfileContext.ACCESS_TOKEN),
            refresh_token=cache.get(CliProfileContext.REFRESH_TOKEN),
            expires_on=cache.get(CliProfileContext.EXPIRES_ON),
            expires_in=cache.get(CliProfileContext.EXPIRES_IN),
            token_type=cache.get(CliProfileContext.TOKEN_TYPE),
            context_name=cache.get(CliProfileContext.NAMESPACE),
            context_type=cache.get(CliProfileContext.CONTEXT_TYPE)
        )

    @active_context.setter
    def active_context(self, context):
        """
        Sets a new active context
        """
        self._profile_cache.set(CliProfileConfiguration.ACTIVE_CONTEXT, context)

    @property
    def contexts(self):
        """
        Gets the contexts available in the profile
        """
        return self._profile_cache.get(CliProfileConfiguration.CONTEXTS)

    @contexts.setter
    def contexts(self, contexts):
        """
        Sets the contexts.
        """
        self._profile_cache.set(CliProfileConfiguration.CONTEXTS, contexts)

    @property
    def accept_eula(self):
        """
        Gets the EULA acceptance.
        """
        return self._profile_cache.get(CliProfileConfiguration.ACCEPT_EULA)

    @accept_eula.setter
    def accept_eula(self, a):
        self._profile_cache.set(CliProfileConfiguration.ACCEPT_EULA, a)

    def get_current_account_user(self):
        return self.active_context.username

    def list(self):
        """
        Get the content of the profile configuration.
        """
        return self._profile_cache.list()

    def persist(self):
        """
        Persist any uncommitted items to the profile. This flushes any modified
        profile items to disk immediately rather than waiting until the end when
        the process finishes.
        """
        self._profile_cache.persist(force=True)


# -----------------------------------------------------------------------------#
# -----------------------------------------------------------------------------#
# -----------------------------------------------------------------------------#


class CliProfileConfiguration(object):
    """
    Manages the `profile.json` cli profile configuration file. This includes
    persisting, deleting, and property lookups.
    """

    ACCEPT_EULA = 'acceptEula'
    CONTEXTS = 'contexts'
    ACTIVE_CONTEXT = 'activeContext'

    def __init__(self, async_persist=None):
        # -- user home --
        cli_name = Configuration().CLI_NAME
        env_profile_override = '{}_CLI_PROFILE'.format(cli_name.upper())
        config_dir = os.path.expanduser(Configuration().CLI_CONFIG_DIR)

        # -- profile --
        ensure_dir(config_dir)
        self._profile = os.path.join(config_dir,
                                     '{}-profile.json'.format(cli_name))

        log.debug(os.environ.get(env_profile_override, None))
        log.debug('profile base location: %s', config_dir)
        log.debug('profile location: %s', self._profile)

        self._uncommitted_items = {}
        self._cached_profile = None
        self._should_flush_to_disk = False
        self._async_persist = async_persist

        if async_persist:
            import atexit
            atexit.register(self.flush_to_disk)

    def persist(self, force=False):
        """
        Persist any uncommitted items to the profile.
        """
        self._should_flush_to_disk = True
        if not self._async_persist or force:
            self.flush_to_disk()

    def flush_to_disk(self):
        # -- flush to disk --
        if self._should_flush_to_disk:

            with os.fdopen(os.open(self._profile,
                                   os.O_RDWR | os.O_CREAT | os.O_TRUNC, 0o600),
                           'w+') as pfile:

                # -- get profile from disk or cache--
                profile = self.list()

                # -- merge in updated `items` for override --
                new_profile = profile.copy()
                new_profile.update(self._uncommitted_items)

                # -- find empty items and remove from new profile --
                empty = {k: v for k, v in new_profile.items() if v is None}
                for k in empty:
                    del new_profile[k]

                # -- write to file --
                pfile.write(json.dumps(new_profile, indent=3))
                self._cached_profile = new_profile
                self._should_flush_to_disk = False

    @property
    def cache(self):
        return self.load_cache()

    def load_cache(self):
        if self._cached_profile is None:
            profile = {}

            # -- load file contents --
            if os.path.isfile(self._profile):
                try:
                    profile = FileUtil.read_json(self._profile) or {}
                except (CliError, ValueError) as ex:
                    raise CliError(
                        "Failed to load token files. Run '{0} login'. "
                        "(Inner Error: {1})".format(Configuration().CLI_NAME,
                                                    ex))

            self._cached_profile = profile

        return self._cached_profile

    def list(self):
        """
        Lists all the key-value pairs in the profile configuration.
        :return:
        """

        self.load_cache()

        return self._cached_profile

    def delete(self, key):
        """
        Deletes the specified `key` in the profile configuration.
        :param key: Key to delete, in the profile configuration.
        :return:
        """
        item = dict()
        item[key] = None
        self._merge(item)
        self.persist()

    def get(self, key):
        """
        Gets the specified `key` in the profile configuration.

        :param key: Key to get in the profile configuration.
        :return: The values of the `key`.
        """

        profile = self.list()

        return profile.get(key)

    def set(self, key, value):
        """
        Sets the specified <key> <value> pair in the profile configuration.

        :param key: Key to set in the profile configuration.
        :param value: Value to set the `key` to.
        """
        item = dict()
        item[key] = value
        self._merge(item)
        self.persist()

    def set_all(self, items):
        self._merge(items)
        self.persist()

    def delete_all(self):
        """
        Removes the entire profile file and all its content.
        """
        try:
            self._uncommitted_items = {}
            self._cached_profile = None
            self._should_flush_to_disk = False

            # we can clear file contents, but deleting it is simpler
            os.remove(self._profile)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise

    def _merge(self, items=dict()):
        """
        Adds items to the uncommited map.
        """
        new_items = self._uncommitted_items.copy()
        new_items.update(items)
        self._uncommitted_items = new_items


_GLOBAL_CLI_PROFILE_CACHE = CliProfileConfiguration(async_persist=True)


# -----------------------------------------------------------------------------#
# -----------------------------------------------------------------------------#
# -----------------------------------------------------------------------------#


def select_authenticator(args):
    """

    :param args:
    :return:
    """
    if 'auth' not in args:
        return NegotiationAuthenticator(**args)

    return {
        SecurityStrategy.basic.name: lambda a: BasicAuthenticator(
            a['client'], a['endpoint'], a['username']),
        SecurityStrategy.ad.name: lambda a: ActiveDirectoryAuthenticator(
            a['client'], a['endpoint'], principal=a.get('principal', None)),
        SecurityStrategy.spnego.name: lambda a: NegotiationAuthenticator(**a)
    }[args['auth']](args)


def acquire_token_by_basic(endpoint, username=None, password=None):
    """
    Convenience function to obtain authentication token by Basic Authentication.

    :param endpoint: The controller endpoint.
    :param username: The username. If not provided it will attempt to lookup the
                     `AZDATA_USERNAME` environment variable if that is not
                     found will prompt for it.
    :param password: The password. If not provided it will attempt to lookup the
                     `AZDATA_PASSWORD` environment variable if that is not
                     found will prompt for it.
    :return: JWT which includes an access-token embedded.
    """
    from azdata.cli.core.clients.cli_client import beget_cli_client

    return select_authenticator({
        'auth': SecurityStrategy.basic.name,
        'client': beget_cli_client(check_auth=False, check_eula=False)(None),
        'endpoint': endpoint,
        'username': username,
        'password': password
    }).authenticate()


def acquire_token_by_ad(endpoint, principal=None):
    """
    Convenience function to obtain authentication token by Active Directory.

    :param endpoint: The controller endpoint.
    :param principal: The optional explicit principal.
    :return: JWT which includes an access-token embedded.
    """
    from azdata.cli.core.clients.cli_client import beget_cli_client

    return select_authenticator({
        'auth': SecurityStrategy.ad.name,
        'client': beget_cli_client(check_auth=False, check_eula=False)(None),
        'endpoint': endpoint,
        'principal': principal,
    }).authenticate()


def acquire_token_by_challenge(endpoint, **args):
    """
    Convenience function to obtain authentication token by a negotiation
    challenge. This helper will attempt to negotiate then authenticate based
    on the controller's service holding security strategy.

    :param endpoint: The controller endpoint.
    :param args: Floating set of arguments based
    :return: JWT which includes an access-token embedded.
    """
    from azdata.cli.core.clients.cli_client import beget_cli_client

    args = {} if not args else args
    args.update({
        'auth': SecurityStrategy.spnego.name,
        'client': beget_cli_client(check_auth=False, check_eula=False)(None),
        'endpoint': endpoint
    })

    return select_authenticator(args).authenticate()


# -----------------------------------------------------------------------------#
# -----------------------------------------------------------------------------#
# -----------------------------------------------------------------------------#


@add_metaclass(ABCMeta)
class BaseAuthenticator(object):
    def __init__(self, client, endpoint):
        self._client = client
        self.client.apis.controller.endpoint = endpoint
        self.client.apis.kerberos.endpoint = endpoint

    @property
    def client(self):
        """A handle to all the different remote client references."""
        return self._client

    @property
    def username(self):
        return self._username

    @property
    def endpoint(self):
        return Configuration().host

    @abstractmethod
    def authenticate(self):
        """Subclasses to imply authentication implementation and return JWT."""

# -----------------------------------------------------------------------------#
# -----------------------------------------------------------------------------#
# -----------------------------------------------------------------------------#


class BasicAuthenticator(BaseAuthenticator):
    """
    Authenticate by way of `Basic Authentication`.
    """

    def __init__(self, client, endpoint, username=None, password=None):
        super(BasicAuthenticator, self).__init__(client, endpoint)
        self._username = username
        self._password = password

    def authenticate(self):
        """
        Authenticate by way of `Basic Authentication` returning a JWT on a
        successful credential passage.
        :return: JWT which includes an access-token embedded,
        """
        def _flush_config_on_error():
            default_config = Configuration()
            default_config.username = ''
            default_config.password = ''

        try:
            # -- Acquire username if not already set --
            if not self._username:
                username = os.getenv(AZDATA_USERNAME)
                if not username:
                    try:
                        username = prompt('Username: ')
                    except NoTTYException:
                        raise NoTTYException('Missing {} '
                                             'environment variable in '
                                             'non-interactive environment.'.
                                             format(AZDATA_USERNAME))
                self._username = username

            # -- Acquire password --
            if not self._password:
                password = os.getenv(AZDATA_PASSWORD)
                if not password:
                    try:
                        password = prompt_pass('Password: ')
                    except NoTTYException:
                        raise NoTTYException('Missing {} '
                                             'environment variable in '
                                             'non-interactive environment.'.
                                             format(AZDATA_PASSWORD))
                self._password = password

            config = Configuration()
            config.username = self._username
            config.password = self._password

            self.client.apis.controller.set_authorization_header(
                'Basic',
                config.get_basic_auth_token(strip_prefix=True)
            )

            return self.client.apis.controller.acquire_token()

        except ApiError as e:
            _flush_config_on_error()
            raise ApiError(status=e.status, reason=e.reason)

# -----------------------------------------------------------------------------#
# -----------------------------------------------------------------------------#
# -----------------------------------------------------------------------------#


class NegotiationAuthenticator(BaseAuthenticator):
    """
    Authenticate by way of negotiating the appropriate security strategy
    (SPNEGO).
    """

    def __init__(self, **args):
        super(NegotiationAuthenticator, self).\
            __init__(args['client'], args['endpoint'])

        self._args = args

    def authenticate(self):
        """
        Authenticate by way of SPNEGO.

        :return: JWT which includes an access-token embedded.
        """

        # -- Attempt to authenticate based on existing profile --
        profile = CliProfile()
        if profile.active_context.security_strategy:
            log.debug('Attempting to authenticate based on profile settings.')
            return select_authenticator({
                'auth': profile.active_context.security_strategy,
                'client': self.client,
                'endpoint': profile.active_context.endpoint,
                'username': profile.active_context.username,
                'principal': profile.active_context.username
            }).authenticate()
        else:
            # -- Attempt to authenticate based on security negotiation --
            log.debug('Attempting to authenticate based on SPNEGO.')

            # Pre-flight to Negotiate security type:
            # Clear authorization header in the request in order to receive a
            # www-authenticate response header to let us know the security
            # holdings. Intended to fail.
            data, status, headers = self.client.request(
                '/token/v1'
                'POST',
                header_params={'Authorization', None}
            )

            log.debug(status)
            log.debug(headers)

            self._args['auth'] = self._negotiate_strategy(headers)
            return select_authenticator(self._args).authenticate()

    def _negotiate_strategy(self, headers):
        """
        Extract the appropriate security strategy in the response header.
        """

        auth = headers.get('www-authenticate', None)

        if auth:
            for strategy in [SecurityStrategy.ad.name,
                             SecurityStrategy.basic.name]:
                if strategy in auth.lower():
                    return strategy

        msg = 'Unsupported security strategy "{}"'.format(auth)
        raise ApiError(reason=msg, status=http_status_codes.unauthorized)

# -----------------------------------------------------------------------------#
# -----------------------------------------------------------------------------#
# -----------------------------------------------------------------------------#


class ActiveDirectoryAuthenticator(BaseAuthenticator):
    """
    Authenticate by way of `Active-Directory/Kerberos`

    """
    def __init__(self, client, endpoint, principal=None):
        super(ActiveDirectoryAuthenticator, self).__init__(client, endpoint)

        self._username = self._principal = principal

    def authenticate(self):
        """
        Authenticate by way of `Active-Directory/Kerberos` returning a JWT on a
        successful credential passage.
        :return: JWT which includes an access-token embedded.
        """

        # Handle an explicit principal. Windows is different in that it uses
        # "username@REALM:password" so we attempt to get the password and suffix
        principal = self._principal
        if principal:
            self._username = principal.split(':')[0]
            if is_windows():
                password = os.getenv('AZDATA_PASSWORD')
                if not password:
                    try:
                        password = prompt_pass('Password: ')
                    except NoTTYException:
                        raise NoTTYException('Missing AZDATA_PASSWORD '
                                             'environment variable in '
                                             'non-interactive environment.')
                principal = '{0}:{1}'.format(principal, password)
        else:
            # if no principle then username is the user of the process
            self._username = os.getenv('USERNAME') if is_windows() \
                else os.getenv('USER')

        if not self._username:
            self._username = '<userId>'

        # 1. acquire GSSAPI token based on principal
        kerberos_client = self.client.apis.kerberos
        kerberos_client.principal = principal

        self.client.apis.controller.set_authorization_header(
            'Negotiate', kerberos_client.acquire_gssapi_token()
         )

        return self.client.apis.controller.acquire_token()
